import { Accessor, createMemo, createResource } from "solid-js";
import getSourceMap from "./get-source-map";

export interface StackFrameSource {
  content: string;
  source: string;
  name?: string;
  line: number;
  column: number;
}

export function createStackFrame(
  stackframe: StackFrame,
  isCompiled: () => boolean,
): Accessor<StackFrameSource> {
  const [data] = createResource(
    () => ({
      fileName: stackframe.fileName,
      line: stackframe.lineNumber,
      column: stackframe.columnNumber,
      functionName: stackframe.functionName,
    }),
    async source => {
      if (!source.fileName) {
        return null;
      }
      const response = await fetch(source.fileName);
      if (!response.ok) {
        return null;
      }
      const content = await response.text();
      const sourceMap = await getSourceMap(source.fileName, content);
      return {
        source,
        content,
        sourceMap,
      };
    },
  );

  const info = createMemo<StackFrameSource>(() => {
    const current = data();
    if (!current) {
      return undefined;
    }
    const { source, content, sourceMap } = current;

    
    if (!isCompiled() && source.line && source.column) {
      const result = sourceMap.originalPositionFor({
        line: source.line,
        column: source.column,
      });

      return {
        ...result,
        content: sourceMap.sourceContentFor(result.source),
      };
    }

    return {
      source: source.fileName,
      line: source.line,
      column: source.column,
      name: source.functionName,
      content,
    };
  });

  return info;
}
